package cca.net.jostl;

import java.net.InetAddress;
import java.net.Socket;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.BufferedReader;
import java.io.StringReader;
import java.util.StringTokenizer;
import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.Calendar;
import java.util.Iterator;




/**
 * 
 * Handles connection with OSCAR server using TOC2 protocol<br><br>
 *<u>November 19, 2006</u><br>
 * - added checking for icq users
 *<br><br>
 * <u>November 10, 2006</u><br>
 * - added support for toc2_login<br>
 * - added IM_IN_ENC2<br>
 * - added CLIENT_EVENT2<br>
 * - added BART2<br>
 * - added CHAT_IN_ENC2<br>
 *
 * <br><br>
 *
 * <u>November 7, 2006</u><br>
 * - fixed bug with firing buddy return status
 *    
 * <br><br><br>  
 * Copyright (C) 2006 by Dustin Zimnox <dzimnox@cca-software.com><br><br><br>
 * 
 * 
 * This library is free software; you can redistribute it and/or modify<br>
 * it under the terms of the GNU Lesser General Public License as published<br>
 * by the Free Software Foundation; either version 2.1 of the License, or<br>
 * (at your option) any later version.<br><br>
 *
 * This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the<br>
 * GNU Lesser General Public License for more details.<br><br>
 *
 * You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA<br>
 *
 * 
 * 
 * @author Dustin Zimnox
 * 
 * @created October 23, 2006
 */



 
public class TOCConnection implements Runnable, TOCConstants
{
    private boolean debug = false;
    
    //protected String loginServer = "toc.oscar.aol.com";
    protected String loginServer = "aimexpress.oscar.aol.com";
    protected int loginPort = 5190;
    protected String authorizerServer = "login.oscar.aol.com";
    protected int authorizerPort = 29999;

    protected final int FLAP_VERSION = 1;
    
    private Socket connection = null;
    private DataInputStream in = null;
    private DataOutputStream out = null;
    
    private final byte FRAME_TYPE_SIGNON = 1;
    private final byte FRAME_TYPE_DATA = 2;
    private final byte FRAME_TYPE_ERROR = 3;
    private final byte FRAME_TYPE_SIGNOFF = 4;
    private final byte FRAME_TYPE_KEEP_ALIVE = 5;

    protected final String REVISION = "\"TIC://Revision: 1.61 \" 160 US \"\" \"\" 3 0 30303 -kentucky -utf8 94791632";
    public String VERSION = "CCTOC";
    protected static final String DEFAULT_GROUP = "Default";
    protected static final String AUTO_ADD_GROUP = "Recent Buddies";
    private int permitMode = 1;

    private int seqNo = (int)Math.floor(Math.random() * 65535.0);
    private String username = null;
    private String password = null;


    //stores all buddies
    private Map<String, TOCBuddy> buddyMap = new HashMap<String, TOCBuddy>();    
    //stores only those buddies a user is attempting to add, then removed after success
    private Map<String, TOCBuddy> addMap = new HashMap<String, TOCBuddy>();
    private Thread connectionThread = null;
    private boolean isConnected = false;
    private boolean receivedConfig = false;
    private ArrayList<TOCListener> listeners = new ArrayList<TOCListener>();



    /**
     * Constructors
     */
    public TOCConnection()
    {
    }

    public TOCConnection(String clientName)
    {
        VERSION = clientName;
    }
    
    
    /**
     *
     */
    public void setDebug(boolean debug)
    {
        this.debug = debug;
    }


    /**
     * Attempts to establish a connection with TOC server
     * @return true if connection was established successfully, false if not
     */
    private boolean connectToTOC() throws IOException
    {
        InetAddress[] loginIPs = null;

        loginIPs = InetAddress.getAllByName(loginServer);
        connection = new Socket(loginIPs[0], loginPort);
        connection.setSoTimeout(10000);
        in = new DataInputStream(connection.getInputStream());
        out = new DataOutputStream(new BufferedOutputStream(connection.getOutputStream()));

        if (connection == null || in == null || out == null)
        {
            return false;
        }
        return true;
    }


    /**
     * Sends signon FLAP
     * 
     * @return true if authorization succeeded, false if not
     */
    private boolean sendSignonFlap()
    {
        try
        {
            out.writeBytes("FLAPON\r\n\r\n");
            out.flush();
            
            //6 byte header, 4 byte FLAP version
            byte[] signonBytes = new byte[10];
            //TOC sends Client FLAP SIGNON
            in.readFully(signonBytes);
            //Client sends TOC FLAP SIGNON
            /* FLAP Header */
            out.writeByte(42); // *
            out.writeByte(FRAME_TYPE_SIGNON);
            out.writeShort(nextSeqNo());
            out.writeShort(username.length()+8); //data length = 8 + username length
            /* SIGNON Frame */
            out.writeInt(FLAP_VERSION);
            out.writeShort(1); //TLV Tag
            String nUsername = normalizeString(username);
            out.writeShort(nUsername.length()); //normalized username length
            out.writeBytes(nUsername);
            out.flush();      
            
            
            /*
            //Client sends TOC "toc2_signon" message
            sendFlap("toc2_signon "+authorizerServer+" "+authorizerPort+" "+nUsername+" "+
                roast(password)+" English "+"\"TIC:"+VERSION+"\" 160 "+magicNumber(nUsername, password)+"\0", FRAME_TYPE_DATA);
            */
            

            //The new, preferred way to sign on
            //This allows for new abilities like talk to icq users
            //incoming messages use IM_IN_ENC2
            //other new commands seen will be: BART2 and BUDDY_CAPS2
            sendFlap("toc2_login "+authorizerServer+" "+authorizerPort+" "+nUsername+" "+
                roast(password)+" English "+REVISION+" "+magicNumber(nUsername, password)+"\0", FRAME_TYPE_DATA);
                         
            
            
            //if login fails TOC drops Client's connection
            //else TOC sends "toc_signon" message
            in.skipBytes(4); //skip frame type and seqNo
            int length = in.readShort(); //length of data
            signonBytes = new byte[length];
            in.readFully(signonBytes);

            String tocResponse = new String(signonBytes);
            if (debug) System.err.println(tocResponse);
            if(tocResponse.equals("SIGN_ON:TOC2.0"))
            {
                sendFlap("toc_init_done\0", FRAME_TYPE_DATA);
                isConnected = true;
                fireConnected();
                
                setCaps(CAP_TALK+" "+
                        CAP_FILE_TRANSFER+" "+
                        CAP_DIRECT_IM+" "+
                        CAP_BUDDY_ICON+" "+
                        CAP_ADD_INS+" "+
                        CAP_FILE_SHARE+" "+
                        CAP_SEND_LIST+" "+
                        CAP_AIMICQ+" "+
                        CAP_ICQ_RELAY+" "+
                        CAP_UTF8);
                
                return true;
            }
            else if(tocResponse.startsWith("ERROR"))
            {
                signOff();
                closeConnection();
            }
        }
        catch(Exception e)
        {}
        return false;
    }



  	/**
     * Gets next number in sequence
     */
    private int nextSeqNo()
    {
        seqNo = (seqNo + 1) & 65535;
        return seqNo;
    }



    /**
     * Processing method for the connection with TOC server.
     */
    public void run()    
    {     
        byte[] data;
        int length;
        byte type;
        
        while(isConnected)
        {
            try
            {
                //check if any bytes are available
                if (in.available() > 0)
                {
                    in.skip(1);
                    type = in.readByte();
                    in.skipBytes(2);
                    length = in.readShort();
                    data = new byte[length];
                    in.readFully(data);
                    if (type == FRAME_TYPE_DATA)
                    {
                        parseTocData(data);
                    }   
                }
            }
            catch(InterruptedIOException e)
            {
                //normal exception, happens when there is no data to read
                System.out.println("IOE: "+e);
            }
            catch(IOException e)
            {
                //System.out.println("error: "+e);
                break;
            }
        }
        try
        {
            closeConnection();
        }
        catch(IOException e)
        {
        }
    }


    /**
     * Returns true if this connection is active
     */
    public boolean isConnected()
    {
        return isConnected;
    }
    
    
    /**
     * Signs onto the TOC network using the TOC2.0 protocol
     * 
     * @param user username to sign on with
     * @param pass password to sign on with
     * @see signOff()
     */
    public void signOn(String user, String pass) throws IOException
    {
        if (!isConnected)
        {
            if (user != null && pass != null)
            {
                this.username = user;
                this.password = pass;
                
                if (connectToTOC())
                {
                    if (sendSignonFlap())
                    {
                        connectionThread = new Thread(this);
                        connectionThread.start();
                    }
                    else
                        throw new IOException("Authorization error");
                }
                else
                     throw new IOException("Error connecting to TOC server");
            }
        }
        else
        {
            throw new IOException("You are already connected");   
        }
    }



    /**
     * Disconnects from the TOC server
     * 
     * @see signOn(String, String)
     */
    public void signOff() throws IOException
    {
        //stops the thread from running
        isConnected = false;
        receivedConfig = false;
        username = null;
        password = null;
        buddyMap.clear();
    }
    

    
    /**
     * Closes the connection and input/output streams
     */
    private void closeConnection() throws IOException
    {
        connection.shutdownInput();
        connection.shutdownOutput();
        connection.close();
        connectionThread = null;
    }


	  
    /**
     * Data received from TOC
     * 
     * @param buffer byte array of data sent from server
     */
    protected void parseTocData(byte[] buffer)
    {
        if (buffer == null)
            return;
        String data = new String(buffer);

        if (debug)
            System.err.println(data);
            
        StringTokenizer token = new StringTokenizer(data,":");
        String command = token.nextToken();
        
        if (command.equals("CONFIG2"))
        {
            if (token.hasMoreTokens())
            {
                String config = token.nextToken();
                while(token.hasMoreTokens())
                {
                    config = config + ":"+token.nextToken();
                }
                processConfig(config);
                fireConfigReceived();
            }
            return;
        }
        
        if (command.equals("IM_IN2"))
        {
            String buddyName = "";
            String auto = "";
            String message = "";
            if (token.hasMoreTokens())
            {
                buddyName = token.nextToken();
                auto = token.nextToken();
                String unknown = token.nextToken();
                while (token.hasMoreTokens())
                    message += token.nextToken();
            }
            
            TOCBuddy buddy = getBuddy(buddyName);
            boolean isBuddy = true;
            
            //buddy is not on list
            if (buddy == null)
            {
                buddy = new TOCBuddy(buddyName, null);
                isBuddy = false;
            }

            fireMessageReceived(buddy, message.trim(), isBuddy, (auto.toUpperCase().equals("T")));
            
            return;
        }
                
        if (command.equals("IM_IN_ENC2"))
        {
            String buddyName    = token.nextToken();
            String auto         = token.nextToken();
            String unknown1     = token.nextToken();
            String unknown2     = token.nextToken();
            String status       = token.nextToken();
            String unknown3     = token.nextToken();
            String unknown4     = token.nextToken();
            String language     = token.nextToken();
            String message      = token.nextToken();
            
            //get rest of message (it might contain ':' that we 
            //dont want to ignore as delimiters)
            if (token.hasMoreTokens())
            {
                while (token.hasMoreTokens())
                    message += token.nextToken();
            }
            
            TOCBuddy buddy = getBuddy(buddyName);
            boolean isBuddy = true;
            
            //buddy is not on list            
            if (buddy == null)
            {
                buddy = new TOCBuddy(buddyName, null);
                isBuddy = false;
            }
            
            fireMessageReceived(buddy, message.trim(), isBuddy, (auto.toUpperCase().equals("T")), language);

            return;
        }       
                
        if (command.equals("CLIENT_EVENT2"))
        {
            String buddyName = token.nextToken();
            int status = Integer.parseInt(token.nextToken());
            
            TOCBuddy buddy = getBuddy(buddyName);
            
            fireClientEvent(buddy, status);
        }
        
        if (command.equals("BART2"))
        {
            TOCBuddy buddy = getBuddy(token.nextToken());
            String unknown = "";
            if (token.hasMoreTokens())
                token.nextToken();
            
            fireBart(buddy, unknown);
        }
       
        //haven't seen this command yet to verify
        if (command.equals("NEW_BUDDY_REPLY2"))
        {
            String b = token.nextToken();
            String action = token.nextToken();
            
            if (action.equals("added"))
            {
                TOCBuddy buddy = addMap.get(normalizeString(b));
                if (buddy != null)
                {
                    buddyMap.put(normalizeString(b),buddy);
                    fireBuddyAdded(buddy);
                    addMap.remove(normalizeString(b));
                }
            }
        }
        
        if (command.equals("INSERTED2"))
        {
            String b = token.nextToken();
            if (b.equals("b"))
            {
                String name = token.nextToken();
                String group = token.nextToken();
                TOCBuddy buddy = new TOCBuddy(name, group);
                buddyMap.put(normalizeString(name), buddy);
                fireBuddyAdded(buddy);
            }
            //else if (b.equals("25"))
        }
        
        if (command.equals("UPDATE_BUDDY2"))
        {
            String b = token.nextToken();
            TOCBuddy buddy = getBuddy(b);
                        
            if (buddy == null) return;
            
            String onlineStatus = token.nextToken();

            int warnAmount = Integer.parseInt(token.nextToken());

            buddy.setWarningAmount(warnAmount);
         
            if (onlineStatus.equals("T"))
            {
                long signOnTime = Long.parseLong(token.nextToken());
                int idleTime = Integer.parseInt(token.nextToken());                               
               
                buddy.setIdleTime(idleTime);
                buddy.setSignOnTime(signOnTime);
                
                String UC = token.nextToken(); //User Class
                
                //notify listeners
                if (!buddy.isOnline())
                {
                    buddy.setOnline(true);
                    fireBuddySignedOn(buddy);
                }

                //check available flag
                if (UC.length() == 3)
                {
                    if (UC.charAt(2) == 'U')
                    {
                        if (!buddy.isAway())
                        {
                            buddy.setAway(true);
                            fireBuddyAway(buddy);
                        }
                    }
                }
                else
                {                    
                    if (buddy.isAway())
                    {                        
                        buddy.setAway(false);
                        fireBuddyReturned(buddy);
                    }
                }
                if (UC.charAt(0) == 'A' || UC.charAt(1) == 'A')
                    buddy.setService(TOCConstants.SERVICE_AOL);
                if (UC.charAt(1) == 'C')
                    buddy.setService(TOCConstants.SERVICE_CELL);
                if (UC.charAt(1) == 'U')
                    buddy.setService(TOCConstants.SERVICE_UNCONFIRMED);
                if (UC.charAt(1) == 'O')
                    buddy.setService(TOCConstants.SERVICE_NORMAL);
                if (UC.charAt(1) == 'I')
                    buddy.setService(TOCConstants.SERVICE_ICQ);
            }
            else if(onlineStatus.equals("F"))
            {
                if (buddy.isOnline())
                {
                    buddy.setOnline(false);
                    fireBuddySignedOff(buddy);
                }
            }
            return;
        }
        
        if (command.equals("ERROR"))
        {
            int error = -1;
            if (token.hasMoreTokens()) 
                error = Integer.parseInt(token.nextToken());
            String args = "";
            while (token.hasMoreTokens())
                args += token.nextToken();
            fireTOCError(error, args.split(":"));
            return;
        }
        
        if (command.equals("EVILED"))
        {
            String amount = token.nextToken();
            TOCBuddy buddy = getBuddy(token.nextToken());
            fireWarning(buddy, amount);
            return;
        }
        
        if (command.equals("CHAT_JOIN"))
        {
            String chatroomID = token.nextToken();
            String chatroom = token.nextToken();
            fireChatJoined(chatroomID, chatroom);
            return;
        }
        
        if (command.equals("CHAT_IN"))
        {
            String chatroomID = token.nextToken();
            String sender = token.nextToken();
            String whisper = token.nextToken(); //T or F
            String message = "";
            while (token.hasMoreTokens())
                    message += token.nextToken();
            fireChatMessageReceived(chatroomID, sender, message, whisper);
            return;
        }

        if (command.equals("CHAT_IN_ENC2"))
        {
            String chatroomID = token.nextToken();
            String sender = token.nextToken();
            String whisper = token.nextToken(); //T or F
            String unknown = token.nextToken();
            String language = token.nextToken();
            String message = token.nextToken();
            while (token.hasMoreTokens())
                    message += token.nextToken();
            fireChatMessageReceived(chatroomID, sender, message, whisper, language, unknown);
            return;
        }
        
        if (command.equals("CHAT_UPDATE_BUDDY"))
        {
            String chatroomID = token.nextToken();
            String inside = token.nextToken(); //T or F
            while(token.hasMoreTokens())
                fireChatUpdate(chatroomID, inside, token.nextToken());
            return;
        }
        
        if (command.equals("CHAT_INVITE"))
        {
            String chatroom = token.nextToken();
            String chatroomID = token.nextToken();
            String sender = token.nextToken();
            String message = "";
            while (token.hasMoreTokens())
                    message += ":"+token.nextToken();
            fireChatInvite(chatroomID, chatroom, sender, message);
            return;
        }
        
        if (command.equals("CHAT_LEFT"))
        {
            String chatroomID = token.nextToken();
            fireChatDropped(chatroomID);
            return;
        }
        
        if (command.equals("GOTO_URL"))
        {
            String windowName = token.nextToken();
            String url = token.nextToken();
            fireUrlReceived(windowName, url);
            return;
        }
        
        if (command.equals("DIR_STATUS"))
        {
            int code = Integer.parseInt(token.nextToken());
            String args = "";
            while (token.hasMoreTokens())
                args += ":"+token.nextToken();
            fireDirStatus(code, args.split(":"));
            return;
        }
        
        if (command.equals("ADMIN_NICK_STATUS"))
        {
            int code = Integer.parseInt(token.nextToken());
            String args = "";
            while (token.hasMoreTokens())
                args += ":"+token.nextToken();
            fireNickStatus(code, args.split(":"));
            return;
        }
        
        if (command.equals("ADMIN_PASSWD_STATUS"))
        {
            int code = Integer.parseInt(token.nextToken());
            String args = "";
            while (token.hasMoreTokens())
                args += ":"+token.nextToken();
            firePasswordStatus(code, args.split(":"));
            return;
        }
        
        if (command.equals("PAUSE"))
        {
            firePause(true);
            return;
        }
        
        if (command.equals("SIGN_ON:TOC2.0"))
        {
            firePause(false);
            return;
        }
        
        if (command.equals("RVOUS_PROPOSE"))
        {
            String name = token.nextToken();
            TOCBuddy buddy = getBuddy(name);
            if (buddy == null) buddy = new TOCBuddy(name, DEFAULT_GROUP);
            
            String UUID = token.nextToken();
            String cookie = token.nextToken();
            int seq = Integer.parseInt(token.nextToken());
            String rip = token.nextToken();
            String pip = token.nextToken();
            String vip = token.nextToken();
            String port = token.nextToken();
            String tlv = "";
            while (token.hasMoreTokens())
                tlv += ":"+token.nextToken();
            
            fireRvousPropose(buddy, UUID, cookie, seq, rip, pip, vip, port, tlv);
            return;
        }
        
        if (command.equals("DELETED2"))
        {
            String type = token.nextToken();
            
            if (type.equals("g"))
            {
				String group = token.nextToken();
				//a group was deleted
            }
            else if(type.equals("b"))
            {
				String buddy = token.nextToken();
                                fireBuddyRemoved(buddy);
				//String group = token.nextToken();
				//a buddy was deleted
            }
            else if(type.equals("d"))
            {
				String buddy = token.nextToken();
				//a buddy was removed from the deny list
            }
            else if(type.equals("p"))
            {
				String buddy = token.nextToken();
				//a buddy was removed from the permit list
            }
            
            return;
        }        
    }

    

    
    /**
     * Parses the config string.
     * The config contains the buddy list information
     * 
     * @param config the config sent by the server
     * @see hasConfig()
     * @see getBuddyList()
     */
    protected void processConfig(String config)
    {
        BufferedReader reader = new BufferedReader(new StringReader(config));
        String line;
        String group = DEFAULT_GROUP;
        try
        {
            while ((line = reader.readLine()) != null)
            {
                TOCBuddy buddy = null;
                if (line.equals("done:"))
                    break;
                
                char type = line.charAt(0);
                String value = (line.substring(2).split(":"))[0];
                
                switch (type)
                {
                    case 'g':
                        group = value;
                        break;
                    case 'b': 
                        buddy = buddyMap.get(normalizeString(value));
                        if (buddy == null)
                        {
                            buddy = new TOCBuddy(value, group);
                            buddyMap.put(normalizeString(value),buddy);
                        }
                        else //buddy already exists
                        {
                            //duplicate buddy, but in different group
							//to allow the list to support duplicate buddies
							//the hashmap key would have to be changed to 
							//something more than just the buddy's name,
							//such as the name + group combined.
                            buddy.setGroup(group);
                        }
                        break;
                    case 'd':
                        buddy = new TOCBuddy(value,group);
                        buddy.setPermissionState(TOCBuddy.PERMISSION_DENY);
                        buddyMap.put(normalizeString(value),buddy);                        
                        break;
                    case 'p':
                        buddy = new TOCBuddy(value,group);
                        buddy.setPermissionState(TOCBuddy.PERMISSION_PERMIT);
                        buddyMap.put(normalizeString(value),buddy);                        
                        break;
                    case 'm':
                        setPermitMode(Integer.parseInt(value));
                        break;                    
                } 
            }
        }
        catch(IOException e)
        {
        }
        catch(NullPointerException ee)
        {
        }
    }


    
    /**
     * Sends raw data without any header
     */
    private void sendRaw(String data) throws IOException
    {
        out.writeBytes(data);
        out.flush();
    }

    

    /**
     * Set's the away message
     *
     *@param message the away message
     */
    public void setAway(String message) throws IOException
    {
        sendFlap("toc_set_away \""+message+"\"\0", FRAME_TYPE_DATA);
    }
    
    
    
    /**
     * Return from away status
     */
    public void returnFromAway() throws IOException
    {
        sendFlap("toc_set_away "+"\0", FRAME_TYPE_DATA);
    }
    
    
    

    /**
     * Lets the server know what capabilities this client supports.
     * Multiple capabilities can be sent by seperating each UUID
     * with a space.
     * 
     * @param caps UUIDs indicating available capabilities supported
     * @see TOCConnection#CAP_TALK
     * @see TOCConnection#CAP_FILE_TRANSFER
     * @see TOCConnection#CAP_DIRECT_IM
     * @see TOCConnection#CAP_BUDDY_ICON
     * @see TOCConnection#CAP_ADD_INS
     * @see TOCConnection#CAP_FILE_SHARE
     * @see TOCConnection#CAP_SEND_LIST
     */
    public void setCaps(String caps) throws IOException
    {
        sendFlap("toc_set_caps "+caps+"\0", FRAME_TYPE_DATA);
    }



    /** 
	 * Gets an TOCBuddy object from a buddy name
     * 
     * @param name buddy name to retrieve
     * @return TOCBuddy
     * @see TOCBuddy
     */    
    public TOCBuddy getBuddy(String name)
    {
        return buddyMap.get(normalizeString(name));
    }


    
    /**
     * Sends a string of data to the server
     * 
     * @param data data to be sent
     * @param frameType FLAP frame type
     * @see TOCConnection#FRAME_TYPE_SIGNON
     * @see TOCConnection#FRAME_TYPE_DATA
     * @see TOCConnection#FRAME_TYPE_ERROR
     * @see TOCConnection#FRAME_TYPE_SIGNOFF
     * @see TOCConnection#FRAME_TYPE_KEEP_ALIVE
     */
    protected void sendFlap(String data, byte frameType) throws IOException
    {
        //FLAP header
        out.writeByte(42); // *
        out.writeByte(frameType);
        out.writeShort(nextSeqNo());
        out.writeShort(data.length());
        //the data to send
        out.writeBytes(data);
        
        out.flush();
    }



    /**
     * TOC2 magic number
     * 
     * @param user username of account to be signed on
     * @param pass password of account to be signed on
     * @return magic number
     */
    private int magicNumber(String user, String pass)
    {
        int sn = user.charAt(0) - 96;
        int pw = pass.charAt(0) - 96;
        int a = sn * 7696 + 738816;
        int b = sn * 746512;
        int c = pw * a;
        return c - a + b + 71665152;
    }
    


    /**
     * Roasts the given string with "Tic/Toc".
     * Roasting a weak form of encrypting the string
     * 
     * @param pass the string to roast (password)
     * @return roasted string in hex format
     */
    private String roast(String pass)
    {
        String roastString = "Tic/Toc";
        String temp = "";
        String out = "0x";
        for (int i=0;i<pass.length();i++)
        {
            temp = java.lang.Long.toHexString(pass.charAt(i) ^ roastString.charAt(i%7));
            if (temp.length() < 2)
                temp = "0"+temp;
            out = out + temp;
        }
        return out;
    }




    /**
     * Normalizes a string
     * 
     * @param s The string to normalize
     * @return the normalized string
     */
    protected static String normalizeString(String s)
    {
        s = s.toLowerCase();
        String str = "";
        char[] chars = s.toCharArray();
        for (int i=0;i<chars.length;i++)
        {
            if (chars[i] != ' ')
            {
                str = str + "" + chars[i];
            }
        }
        return str;
    }



    /**
     * Returns an Iterator containing all TOCBuddy's created from
     * the CONFIG2 sent by TOC
     * 
     * @return Iterator
     * @see hasConfig()
     */
    public Iterator getBuddyList()
    {
        return buddyMap.values().iterator();
    }



    /**
     * Sets the permit mode
     * Use the static permit mode variables
     * 
     * @param p permit mode
     * @see TOCConnection#PERMIT_ALL
     * @see TOCConnection#DENY_ALL
     * @see TOCConnection#PERMIT_SOME
     * @see TOCConnection#DENY_SOME
     * @see TOCConnection#ALLOW_BUDDIES_ONLY
     */
    public void setPermitMode(int p)
    {
        permitMode = p;
    }



    /**
     * Adds a new buddy to the user's buddy list
     * If the group the buddy belongs to doesn't
     * already exist, it will be created.
     * 
     * @param buddy TOCBuddy to add
     * @see TOCBuddy
     */
    public void addBuddy(TOCBuddy buddy) throws IOException
    {

        if (buddy == null)
            return;

        if (getBuddy(buddy.getName()) != null)
            return;

        String data = "toc2_new_buddies {g:"+buddy.getGroup()+"\nb:"+buddy.getName()+"\n}";
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
        
        addMap.put(normalizeString(buddy.getName()), buddy);
    }
    
    

    /**
     * Adds a new buddy to the user's buddy list
     * If the group the buddy belongs to doesn't
     * already exist, it will be created.
     * 
     * @param buddy name of buddy to add
     * @param group name of group the buddy belongs to
     */
    public void addBuddy(String name, String group) throws IOException
    {

        if (name == null)
            return;

        if (getBuddy(name) != null)
            return;
        
        if (group == null)
            group = DEFAULT_GROUP;

        String data = "toc2_new_buddies {g:"+group+"\nb:"+name+"\n}";
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
        
        TOCBuddy buddy = new TOCBuddy(name, group);
        addMap.put(normalizeString(name), buddy);
    }
 

    /**
     * Sends an instant message to a buddy.
	 * Messages greater than MESSAGE_LENGTH_LIMIT will 
	 * still be sent but only a substring of the first X
	 * number of characters will be used.
     * 
     * @param buddy name of buddy to send message to
     * @param message text to send to buddy, including HTML
	 * @see TOCConstants#MESSAGE_LENGTH_LIMIT
     */
    public void sendMessage(String buddy, String message) throws IOException
    {
        if (message.length() >= MESSAGE_LENGTH_LIMIT)
            message = message.substring(0,MESSAGE_LENGTH_LIMIT);

        String data = "toc2_send_im "+normalizeString(buddy)+" \""+message+"\"";
        
        sendFlap(data+"\0", FRAME_TYPE_DATA);
    }
    


    /**
     * Requests status of a buddy
     * 
     * @param buddy buddy to get status of
     */
    private void getBuddyStatus(TOCBuddy buddy) throws IOException
    {
        if (buddy == null)
            return;            
        
        String data = "toc_get_status "+normalizeString(buddy.getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }



    /**
     * Removes a buddy from the user's buddy list
     * 
     * Currently not working??????
     * 
     * @param buddy TOCBuddy to remove
     * @see TOCBuddy
     */
    public void removeBuddy(TOCBuddy buddy) throws IOException
    {
        if (buddy == null)
            return;

        String data = "toc2_remove_buddy "+normalizeString(buddy.getName())+" \""+buddy.getGroup()+"\"";
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }



    /**
     * Sends an invitation to a buddy
     * 
     * @param roomId ID of the chatroom to invite buddy to
     * @param buddy TOCBuddy to invite
     * @param message greeting message
     * @see TOCBuddy
     * @see createRoom(String) 
     */
    public void sendChatInvite(String roomId, TOCBuddy buddy, String message) throws IOException
    {
        if (buddy == null)
            return;
            
        String data = "toc_chat_invite "+roomId+" \""+message+"\" "+normalizeString(buddy.getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    


    /**
     * Sends an invitation to a buddy
     * 
     * @param roomId ID of the chatroom to invite buddy to
     * @param buddy name of buddy to invite
     * @param message greeting message
     * @see createRoom(String) 
     */
    public void sendChatInvite(String roomId, String buddy, String message) throws IOException
    {
        if (buddy == null)
            return;
            
        String data = "toc_chat_invite "+roomId+" \""+message+"\" "+normalizeString(buddy);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }   



    /**
     * Creates a new chat room.  Upon successful
     * creation, the server will send CHAT_JOIN
     * followed by a CHAT_UPDATE_BUDDY about the
     * user itself. 
     * 
     * @param name the name for the chatroom
     * @see TOCListener#chatroomJoined(String, String)
     * @see TOCListener#chatroomUpdate(String, String, boolean)
     */
    public void createChatRoom(String name) throws IOException
    {
        String data = "toc_chat_join 4 "+name;

        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }



    /**
     * Sends a message to the chatroom readable by everyone in that room.
     * User will receive this message from the server along with all other
     * users in the room.
     *
     * @param roomID ID of the chat room
     * @param message the message, including any html tags, to send to the chatroom
     * @see TOCListener#chatroomMessageReceived(String, String, String, boolean)
     */
    public void sendChatMessage(int roomID, String message) throws IOException
    {
        String data = "toc_chat_send " + roomID + " \""+message+"\"";
        System.out.println("Connection: "+data);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    


    /**
     * Whisper a message to only a specific buddy in the chatroom
     *
     * @param roomID ID of the chatroom
     * @param buddy TOCBuddy to send private message to
     * @param message The message to send
     * @see TOCBuddy
     */
    public void sendChatWhisper(int roomID, TOCBuddy buddy, String message) throws IOException
    {
        String data = "toc_chat_whisper " + roomID + " "+normalizeString(buddy.getName()) + " \"" + message+"\"";
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }



    /**
     * Whisper a message to only a specific buddy in the chatroom
     *
     * @param roomID ID of the chatroom
     * @param buddy name of buddy to send private message to
     * @param message The message to send
     */
    public void sendChatWhisper(int roomID, String buddy, String message) throws IOException
    {
        String data = "toc_chat_whisper " + roomID + normalizeString(buddy) + "\"" + message+"\"";
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    
    

    /**
     * Warns a buddy in the chatroom
     *
     * @param roomID ID of the chatroom
     * @param buddy TOCBuddy to warn
     * @param anonymous true to warn anonymously, false to show buddy your indentity
     * @see TOCBuddy
     */
    public void sendChatWarning(int roomID, TOCBuddy buddy, boolean anonymous) throws IOException
    {   
        String from = "norm";
        if (anonymous)
            from = "anon";
            
        String data = "toc_chat_evil " + roomID + normalizeString(buddy.getName()) + from;
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Warns a buddy in the chatroom
     *
     * @param roomID ID of the chatroom
     * @param buddy name of buddy to warn
     * @param anonymous true to warn anonymously, false to show buddy your indentity
     */
    public void sendChatWarning(int roomID, String buddy, boolean anonymous) throws IOException
    {   
        String from = "norm";
        if (anonymous)
            from = "anon";
            
        String data = "toc_chat_evil " + roomID + normalizeString(buddy) + from;
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    
    

    /**
     * Leaves the specified chatroom
     * 
     * @param roomID ID of the chatroom 
     */
    public void leaveChat(int roomID) throws IOException
    {
        String data = "toc_chat_leave "+roomID;
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Accepts a chatroom invitation. Upon success, client
     * will receive the chatroomJoined event.
     * 
     * @param roomID ID of the chat room
     * @see TOCListener#chatroomJoined(String, String)
     */
    public void acceptChatInvite(int roomID) throws IOException
    {
        String data = "toc_chat_accept "+roomID;
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    
    
        
    /**
     * Sets the user idle. Server handles updating 
     * the idle time, so subsequent calls are not
     * necessary.
     * 
     * @param seconds current idle time in seconds
     */
    public void setIdle(int seconds) throws IOException
    {
        String data = "toc_set_idle "+seconds;
        sendFlap(data+"\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Sets the users status back from being idle.
     */
    public void returnFromIdle() throws IOException
    {
        String data = "toc_set_idle 0";
        sendFlap(data+"\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Sets user's profile information
     * 
     * @param info profile info
     */
    public void setInfo(String info) throws IOException
    {
        String data = "toc_set_info \""+info+"\"";

        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    
    

    /**
     * Retrieves a buddy's information in the form of
     * a GOTO_URL sent back from the server
     * 
     * @param buddy TOCBuddy to get info for
     * @see TOCListener#urlReceived(String, String)
     */
    public void getBuddyInfo(TOCBuddy buddy) throws IOException
    {
        String data = "toc_get_info "+normalizeString(buddy.getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Retrieves a buddy's information in the form of
     * a GOTO_URL sent back from the server
     * 
     * @param buddy name of buddy to get info for
     * @see TOCListener#urlReceived(String, String)
     */
    public void getBuddyInfo(String buddy) throws IOException
    {
        String data = "toc_get_info "+normalizeString(buddy);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    


    /**
     *  Retrieves a buddy's directory info
     *
     * @param buddy TOCBuddy to retrieve directory info for
     * @see TOCListener#urlReceived(TOCBuddy, String)
     */
    public void getDirInfo(TOCBuddy buddy) throws IOException
    {
        String data = "toc_get_dir " + normalizeString(buddy.getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     *  Retrieves a buddy's directory info
     *
     * @param buddy buddy name to retrieve directory info for
     * @see TOCListener#urlReceived(TOCBuddy, String)
     */
    public void getDirInfo(String buddy) throws IOException
    {
        String data = "toc_get_dir " + normalizeString(buddy);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }



    /**
     * Sets the user's directory information.
     * You should receive a dirStatus event following this method call
     * 
     * @param first your first name
     * @param middle your middle name
     * @param last your last name
     * @param maiden your maiden name
     * @param city your city
     * @param state your state
     * @param country your country
     * @param email your email
     * @param webSearch Having anything in this field allows others
     *      to find your dir info from web-searches, else they have 
     *      to use the client
     * @see TOCListener#dirStatus(int, String[])
     */
    public void setDirInfo(String first, String middle, String last, String maiden,
                            String city, String state, String country, String email,
                            String webSearch) throws IOException
    {
        String data = "toc_set_dir \"" + first + "\":\"" + middle + "\":\"" + last + 
                      "\":\"" + maiden + "\":\"" + city + "\":\"" + state + "\":\"" + 
                      country + "\":\"" + email + "\":\"" + webSearch + "\"";
                      
        sendFlap(data + "\0", FRAME_TYPE_DATA);              
    }



    /**
     * Sets the user's directory information.
     * You will either receive an aimError or urlReceived
     * 
     * @param first your first name
     * @param middle your middle name
     * @param last your last name
     * @param maiden your maiden name
     * @param city your city
     * @param state your state
     * @param country your country
     * @param email your email
     * @param webSearch Search using other keywords
     * @see TOCListener#aimError(int, String[])
     * @see TOCListener#urlReceived(String, String)
     */
    public void searchDirectory(String first, String middle, String last, String maiden,
                            String city, String state, String country, String email,
                            String webSearch) throws IOException
    {
        String data = "toc_dir_search \"" + first + "\":\"" + middle + "\":\"" + last + 
                      "\":\"" + maiden + "\":\"" + city + "\":\"" + state + "\":\"" + 
                      country + "\":\"" + email + "\":\"" + webSearch + "\"";
                      
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }


  
    /**
     * Creates a new group on the user's buddy list
     * 
     * @param group name of the group to add
     */
    public void createGroup(String group) throws IOException
    {
        String data = "toc2_new_group \""+group+"\"";
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    
    
    /**
     * Deletes a group on the user's buddy list
     * 
     * @param group name of the group to delete
     */
    public void deleteGroup(String group) throws IOException
    {
        String data = "toc2_del_group \""+group+"\"";
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }


    
    /**
     * Adds a buddy to the deny list
     * 
     * @param buddy TOCBuddy to put on deny list
     * @see TOCBuddy
     */
    public void addToDeny(TOCBuddy buddy) throws IOException
    {
        String data = "toc2_add_deny "+normalizeString(buddy.getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Adds multiple buddies to the deny list
     * 
     * @param buddies array of TOCBuddy's to put on deny list
     * @see TOCBuddy
     */
    public void addToDeny(TOCBuddy[] buddies) throws IOException
    {
        String data = "toc2_add_deny";
        for (int i=0; i<buddies.length; i++)
            data += " "+normalizeString(buddies[i].getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Adds a buddy to the deny list
     * 
     * @param buddy name to put on deny list
     */
    public void addToDeny(String buddy) throws IOException
    {
        String data = "toc2_add_deny "+normalizeString(buddy);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    
    

    /**
     * Adds multiple buddies to the deny list
     * 
     * @param buddies array of buddy names to put on deny list
     */
    public void addToDeny(String[] buddies) throws IOException
    {
        String data = "toc2_add_deny";
        for (int i=0; i<buddies.length; i++)
            data += " "+normalizeString(buddies[i]);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }



    /**
     * Removes a buddy from the deny list
     * 
     * @param buddy TOCBuddy to remove from deny list
     * @see TOCBuddy
     */
    public void removeFromDeny(TOCBuddy buddy) throws IOException
    {
        String data = "toc2_remove_deny "+normalizeString(buddy.getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
        
    

    /**
     * Removes multiple buddies from the deny list
     * 
     * @param buddies array of TOCBuddy's to remove from deny list
     * @see TOCBuddy
     */
    public void removeFromDeny(TOCBuddy[] buddies) throws IOException
    {
        String data = "toc2_remove_deny";
        for (int i=0; i<buddies.length; i++)
            data += " "+normalizeString(buddies[i].getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    
    

    /**
     * Removes a buddy from the deny list
     * 
     * @param buddy name to remove from deny list
     */
    public void removeFromDeny(String buddy) throws IOException
    {
        String data = "toc2_remove_deny "+normalizeString(buddy);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }

    
    
    /**
     * Removes multiple buddies from the deny list
     * 
     * @param buddies array of buddy names to remove from deny list
     */
    public void removeFromDeny(String[] buddies) throws IOException
    {
        String data = "toc2_remove_deny";
        for (int i=0; i<buddies.length; i++)
            data += " "+normalizeString(buddies[i]);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }

    
    
    /**
     * Adds a buddy to the permit list
     * 
     * @param buddy TOCBuddy to put on permit list
     * @see TOCBuddy
     */
    public void addToPermit(TOCBuddy buddy) throws IOException
    {
        String data = "toc2_add_permit "+normalizeString(buddy.getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }


    
    /**
     * Adds multiple buddies to the permit list
     * 
     * @param buddies array of TOCBuddy's to put on permit list
     * @see TOCBuddy
     */
    public void addToPermit(TOCBuddy[] buddies) throws IOException
    {
        String data = "toc2_add_permit";
        for (int i=0; i<buddies.length; i++)
            data += " "+normalizeString(buddies[i].getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    
    

    /**
     * Adds a buddy to the permit list
     * 
     * @param buddy name to put on permit list
     */
    public void addToPermit(String buddy) throws IOException
    {
        String data = "toc2_add_permit "+normalizeString(buddy);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }


    
    /**
     * Adds multiple buddies to the permit list
     * 
     * @param buddies array of buddy names to put on permit list
     */
    public void addToPermit(String[] buddies) throws IOException
    {
        String data = "toc2_add_permit";
        for (int i=0; i<buddies.length; i++)
            data += " "+normalizeString(buddies[i]);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }

    
    
    /**
     * Removes a buddy from the permit list
     * 
     * @param buddy TOCBuddy to remove from permit list
     * @see TOCBuddy
     */
    public void removeFromPermit(TOCBuddy buddy) throws IOException
    {
        String data = "toc2_remove_permit "+normalizeString(buddy.getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Removes multiple buddies from the permit list
     * 
     * @param buddies array of TOCBuddy's to remove from permit list
     * @see TOCBuddy
     */
    public void removeFromPermit(TOCBuddy[] buddies) throws IOException
    {
        String data = "toc2_remove_permit";
        for (int i=0; i<buddies.length; i++)
            data += " "+normalizeString(buddies[i].getName());
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Removes a buddy from the permit list
     * 
     * @param buddy name to remove from permit list
     */
    public void removeFromPermit(String buddy) throws IOException
    {
        String data = "toc2_remove_permit "+normalizeString(buddy);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * Removes multiple buddies from the permit list
     * 
     * @param buddies array of buddy names to remove from permit list
     */
    public void removeFromPermit(String[] buddies) throws IOException
    {
        String data = "toc2_remove_permit";
        for (int i=0; i<buddies.length; i++)
            data += " "+normalizeString(buddies[i]);
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }



    /**
     * Changes the format of the username
     * This can only change the spacing of the characters, and
     * capitalization of the characters. If both the old format
     * and new format are normalized, they should appear identical.
     * In response, the server will send either ADMIN_NICK_STATUS
     * or ERROR.
     * 
     * @param name new formatted name
     * @see TOCListener#adminFormatStatus(int)
     * @see TOCListener#aimError(int, String[])
     */
    public void setNameFormat(String name) throws IOException
    {
        String data = "toc_format_nickname \""+name+"\"";

        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    
    

    /**
     * Changes the password for the account
     * Passwords cannot contain spaces
     * 
     * @param oldPass old/current password
     * @param newPass new password
     * @see TOCListener#adminPasswordStatus(int)
     * @see TOCListener#aimError(int, String[])
     */
    public void changePassword(String oldPass, String newPass) throws IOException
    {
        String data = "toc_change_passwd "+oldPass+" "+newPass;
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);        
    }


    
    /**
     * Accepts a rendezvous proposal
     * 
     * @param buddy TOCBuddy to accept proposal from
     * @param cookie cookie sent with the initial proposal
     * @param service the UUID of the service this proposal is to be used for
     * @param tlv any tlv's that need to be sent, base64 encoded values
     * @see TOCListener#rvousPropose(TOCBuddy,String,String,String,String,String,String,String,String)
     */
    public void acceptRvous(TOCBuddy buddy, String cookie, String service, String tlv) throws IOException
    {
        String data = "toc_rvous_accept "+normalizeString(buddy.getName())+" "+cookie+" "+service+" "+tlv;
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }



    /**
     * Accepts a rendezvous proposal
     * 
     * @param buddy name of buddy to accept proposal from
     * @param cookie cookie sent with the initial proposal
     * @param service the UUID of the service this proposal is to be used for
     * @param tlv any tlv's that need to be sent, base64 encoded values
     * @see TOCListener#rvousPropose(TOCBuddy,String,String,String,String,String,String,String,String)
     */
    public void acceptRvous(String buddy, String cookie, String service, String tlv) throws IOException
    {
        String data = "toc_rvous_accept "+normalizeString(buddy)+" "+cookie+" "+service+" "+tlv;
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    

    
    /**
     * @param buddy TOCBuddy to cancel proposal from
     * @param cookie cookie sent with the initial proposal
     * @param service the UUID of the service this proposal is to be used for
     * @param tlv any tlv's that need to be sent, base64 encoded values
     * @see TOCListener#rvousPropose(TOCBuddy,String,String,String,String,String,String,String,String)
     */
    public void cancelRvous(TOCBuddy buddy, String cookie, String service, String tlv) throws IOException
    {
        String data = "toc_rvous_cancel "+normalizeString(buddy.getName())+" "+cookie+" "+service+" "+tlv;
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }



    /**
     * @param buddy name of buddy to cancel proposal from
     * @param cookie cookie sent with the initial proposal
     * @param service the UUID of the service this proposal is to be used for
     * @param tlv any tlv's that need to be sent, base64 encoded values
     * @see TOCListener#rvousPropose(TOCBuddy,String,String,String,String,String,String,String,String)
     */
    public void cancelRvous(String buddy, String cookie, String service, String tlv) throws IOException
    {
        String data = "toc_rvous_cancel "+normalizeString(buddy)+" "+cookie+" "+service+" "+tlv;
        
        sendFlap(data + "\0", FRAME_TYPE_DATA);
    }
    


    /**
     * Propose a rendezvous - not yet implemented
     */
    public void proposeRvous(String buddy, String uuid, String cookie, String seq, String rip,
                                String pip, String vip, String port, String tlv) throws IOException
    {
        
        String data = "RVOUS_PROPOSE "+normalizeString(buddy)+" "+uuid+" "+cookie+" "+seq+" "+rip+" "+
                        pip+" "+vip+" "+port+" "+tlv;
        sendFlap(data+"\0", FRAME_TYPE_DATA);

        //method doesn't work
    }



    /**
     * Determines if buddy is on the user's permit list
     * 
     * @param buddy TOCBuddy to check for
     * @return true or false
     * @see TOCBuddy
     */
    public boolean isBuddyOnPermit(TOCBuddy buddy)
    {
        if (buddy.getPermissionState() == TOCBuddy.PERMISSION_PERMIT)
            return true;

        return false;
    }    
    

    
    /**
     * Determines if the buddy is on the user's deny list
     * 
     * @param buddy TOCBuddy to check for
     * @return true or false
     * @see TOCBuddy
     */
    public boolean isBuddyOnDeny(TOCBuddy buddy)
    {
        if (buddy.getPermissionState() == TOCBuddy.PERMISSION_DENY)
            return true;

        return false;
    }



    /**
     * Returns true if the config was received from TOC
     * @return true if connection has received the config
     * @see getBuddyList()
     */
    public boolean hasConfig()
    {
        return receivedConfig;
    }



    /**
     * Returns the permit mode
     * 
     * @see TOCConnection#PERMIT_ALL
     * @see TOCConnection#DENY_ALL
     * @see TOCConnection#PERMIT_SOME
     * @see TOCConnection#DENY_SOME
     * @see TOCConnection#ALLOW_BUDDIES_ONLY
     * 
     * @return permit mode 
     */
    public int getPermitMode()
    {
        return permitMode;
    }
    


    /**
     * Gets the name of the user currently signed in
     * 
     * @return username of the account signed on with this connection
     */
    public String getUsername()
    {
        return username;
    }



    /**
     * Sets the login server
     * Default address: toc.oscar.aol.com
     * Default port: 5190
     * 
     * @param server new server address
     * @param port new server port
     */
    public void setLoginServer(String server, int port)
    {
        loginServer = server;
        loginPort = port;
    }



    /**
     * Sets the authorization server
     * Default address: login.oscar.aol.com
     * Default port: 29999
     * 
     * @param server new server address
     * @param port new server port
     */
    public void setAuthorizationServer(String server, int port)
    {
        authorizerServer = server;
        authorizerPort = port;
    }
  
    

    /**
     * Gets the login server address
     * 
     * @return address string of the login server
     * @see setLoginServer(String, int)
     */
    public String getLoginServerAddress()
    {
        return loginServer;
    }



    /**
     * Gets the login server port
     * 
     * @return port of the login server
     * @see setLoginServer(String, int)
     */
    public int getLoginServerPort()
    {
        return loginPort;
    }
    


    /**
     * Gets the authorization server address
     * 
     * @return address string of the authorization server
     * @see setAuthorizationServer(String, int)
     */
    public String getAuthorizationServerAddress()
    {
        return authorizerServer;
    }
    


    /**
     * Gets the authorization server port
     * 
     * @return port of the authorization server
     * @see setAuthorizationServer(String, int)
     */
    public int getAuthorizationServerPort()
    {
        return authorizerPort;
    }



    /**
     * Register a listener with this TOCConnection to receive events
     * @param listener
     *          The listener  
     */
    public void addTOCListener(TOCListener listener)
    {
        listeners.add(listener);   
    }
    

    
    /**
     * Register a listener with this TOCConnection to receive events
     * @param listener
     *          The listener  
     */
    public void removeTOCListener(TOCListener listener)
    {
        int index = listeners.indexOf(listener);
        if (index > -1)
            listeners.remove(index);
    }












    
    /**
     * Notify listeners that a connection has been established
     */
    private void fireConnected()
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).connected();
        }
    }
    


    /**
     * Notify listeners that the connection was closed
     */
    private void fireDisconnected()
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).disconnected();
        }
    }
    


    /**
     * Notify listeners config was received
     */
    private void fireConfigReceived()
    {
        receivedConfig = true;
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).buddyListReceived(getBuddyList());
        }
    }    
    
    
    
    /**
     * Notify listeners of a buddy art response
     * @param buddy
     * @param unknown
     */
    private void fireBart(TOCBuddy buddy, String unknown)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).buddyArt(buddy, unknown);
        }
    }
    


    /**
     * Notify listeners that a message was recieved from a buddy
     * @param buddy
     *      buddy who sent the message
     * @param message
     *      text from buddy, includes HTML tags
     * @param isBuddy
     *      is true if incoming message is from a buddy on the user's list, else false
     * @param auto
     *      true if this message is an auto response (away message)
     */
    private void fireMessageReceived(TOCBuddy buddy, String message, boolean isBuddy, boolean auto)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).messageReceived(buddy, message, isBuddy, auto);
        }
    }
    

    
    /**
     * Notify listeners that a message was recieved from a buddy
     * @param buddy
     *      buddy who sent the message
     * @param message
     *      text from buddy, includes HTML tags
     * @param isBuddy
     *      is true if incoming message is from a buddy on the user's list, else false
     * @param auto
     *      true if this message is an auto response (away message)
     * @param language
     */
    private void fireMessageReceived(TOCBuddy buddy, String message, boolean isBuddy, boolean auto, String language)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).messageReceived(buddy, message, isBuddy, auto, language);
        }
    }
    
    
    
    /**
     * Notify listeners of client events
     * @param buddy
     *      buddy who this event pertains to
     * @param status
     *      client status of buddy, see TOCConstants
     */
    private void fireClientEvent(TOCBuddy buddy, int status)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).clientEvent(buddy, status);
        }
    }
    
    
    
    /**
     * Notify listeners that a buddy has signed on
     * @param buddy
     *      buddy who signed on
     */
    private void fireBuddySignedOn(TOCBuddy buddy)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).buddySignedOn(buddy);
        }
    }
    


    /**
     * Notify listeners that a buddy as signed off
     * @param buddy
     *      buddy who signed off
     */
    private void fireBuddySignedOff(TOCBuddy buddy)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).buddySignedOff(buddy);
        }
    }


    
    /**
     * Notify listeners that a buddy has set themselves as away
     * @param buddy
     */
    private void fireBuddyAway(TOCBuddy buddy)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).buddyAway(buddy);
        }
    }


    
    /**
     * Notify listeners that a buddy has returned from away status
     */
    private void fireBuddyReturned(TOCBuddy buddy)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).buddyReturned(buddy);
        }
    }
    


    /**
     * Notify listeners that a buddy has been added successfully
     */
    private void fireBuddyAdded(TOCBuddy buddy)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).buddyAdded(buddy);
        }
    }
    


    /**
     *
     */
    private void fireBuddyRemoved(String name)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).buddyRemoved(name);
        }
    }
    
    
    
    /**
     * Notify listeners an error has occurred
     * @param err
     *      error code
     * @param args
     *      additional arguments supplied by error
     */
    private void fireTOCError(int err, String[] args)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).aimError(err, args);
        }
    }
    


    /**
     * Notify listeners that user has joined a chatroom
     * @param chatroomID
     *      ID of chatroom, internal to TOC
     * @param chatroom
     *      name of chatroom
     */
    private void fireChatJoined(String chatroomID, String chatroom)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).chatroomJoined(chatroomID, chatroom);
        }  
    }
    


    /**
     * Notify listeners a message was received in the chatroom
     * @param chatroomID
     *      ID of chatroom, internal to TOC
     * @param sender
     *      name of buddy who sent message (name not normalized)
     * @param message
     *      text from buddy, includes HTML tags
     * @param whisper
     *      if true, only this TOCConnection received this message
     * 
     */
    private void fireChatMessageReceived(String chatroomID, String sender, String message, String whisper)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).chatroomMessageReceived(chatroomID, sender, message, whisper.toUpperCase().equals("T"));
        } 
    }
    
    
    /**
     * Notify listeners a message was received in the chatroom
     * @param chatroomID
     *      ID of chatroom, internal to TOC
     * @param sender
     *      name of buddy who sent message (name not normalized)
     * @param message
     *      text from buddy, includes HTML tags
     * @param whisper
     *      if true, only this TOCConnection received this message
     * @param language
     *      'en' for english
     * @param unknown
     */
    private void fireChatMessageReceived(String chatroomID, String sender, String message, String whisper, String language, String unknown)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).chatroomMessageReceived(chatroomID, sender, message, whisper.toUpperCase().equals("T"), language, unknown);
        }
    }

    
    
    /**
     * Notify listeners about buddys joining and leaving the chatroom
     * @param chatroomID
     *      ID of chatroom, internal to TOC
     * @param entering
     *      true if buddy is entering the room, false if buddy is leaving 
     * @param buddy
     *      name of buddy (name not normalized)
     */
    private void fireChatUpdate(String chatroomID, String entering, String buddy)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).chatroomUpdate(chatroomID, buddy, entering.toUpperCase().equals("T"));
        }
    }
    


    /**
     * Notify listeners of a chatroom invitation
     * @param chatroomID
     *      ID of chatroom, internal to TOC
     * @param chatroom
     *      name of chatroom
     * @param sender
     *      name of buddy who sent the invite (name not normalized)
     * @param message
     *      invite message (includes HTML??)
     */
    private void fireChatInvite(String chatroomID, String chatroom, String sender, String message)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).chatroomInvite(chatroomID, chatroom, sender, message);
        }
    }
    


    /**
     * Notify listeners the chatroom was closed
     * @param chatroomID ID of chatroom, received when user joined chatroom
     */
    private void fireChatDropped(String chatroomID)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).chatroomDropped(chatroomID);
        }
    }
    

    
    /**
     * Notify listeners a GOTO_URL was received from the server
     * 
     * @param windowName suggested internal name for window
     * @param url the url string
     */
    private void fireUrlReceived(String windowName, String url)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).urlReceived(windowName, url);
        }
    }
    
    

    /**
     * Notify listeners a buddy list was received from the server
     * 
     * @param list an Iterator of TOCBuddy objects
     */
    private void fireBuddyListReceived(Iterator list)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).buddyListReceived(list);
        }
    }
    
    

    /**
     * Notify listeners of return code from server in response to an attempt to change the format of the client's name
     * 
     * @param code error code, 0 equals success
     * @param args optional arguments
     */
    private void fireNickStatus(int code, String[] args)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).adminFormatStatus(code, args);
        }
    }
    

    
    /**
     * Notify listeners of return code from server in response to an attempt to change the client's password
     * 
     * @param code error code, 0 equals success
     * @param args optional arguments
     */
    private void firePasswordStatus(int code, String[] args)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).adminPasswordStatus(code, args);
        }
    }
    

    
    /**
     * Notify listeners of return code from server in response to toc_set_dir
     * 
     * @param code error code, 0 equals success
     * @param args optional arguments
     */
    private void fireDirStatus(int code, String[] args)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).dirStatus(code, args);
        }
    }


    
    /**
     * Notify listeners this client was warned
     * 
     * @param buddy TOCBuddy who sent the warning, null if anonymous
     * @param amount The new warning amount
     */
    private void fireWarning(TOCBuddy buddy, String amount)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).warned(buddy, amount);
        }
    }
    
    

    /**
     * Notify listeners that TOC server has either paused or unpaused communication
     * 
     * @param state true for paused, false for unpaused
     */
    private void firePause(boolean state)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).isPaused(state);
        }
    }    
    
    
    
    /**
     * Notify listeners another user has proposed a rendezvous
     */
    private void fireRvousPropose(TOCBuddy buddy, String UUID, String cookie, int seq,
                                    String rip, String pip, String vip, String port, String tlv)
    {
        for(int i=0;i<listeners.size();i++)
        {
           ((TOCListener)listeners.get(i)).rvousPropose(buddy, UUID, cookie, seq, rip, pip, vip, port, tlv);
        }
    }

}
