package cca.net.jostl;

import java.util.Iterator;
import java.util.EventListener;


/**
 * Interface for TOC connection listener
 * 
 * @author Dustin Zimnox
 * 
 * @created October 23, 2006
 * @edited November 10, 2006
 */
public interface TOCListener extends EventListener
{  
    /** Successfully connected */
    void connected();
    
    /** disconnected from server */
    void disconnected();    
    
    /**
     * buddy removed successfully
     * @param name the name of the buddy who was removed
     */
    void buddyRemoved(String name);
        
    /**
     * incoming message
     * DEPRECATED
     * @param buddy who the message is from
     * @param message the actual message, including html
     * @param isBuddy true if this person is on your buddy list
     * @param auto true if this message is an auto response (away message)
     */
    void messageReceived(TOCBuddy buddy, String message, boolean isBuddy, boolean auto);
        
    /**
     * incoming message
     * 
     * @param buddy who the message is from
     * @param message the actual message, including html
     * @param isBuddy true if this person is on your buddy list
     * @param auto true if this message is an auto response (away message)
     * @param language
     */
    void messageReceived(TOCBuddy buddy, String message, boolean isBuddy, boolean auto, String language);
    
    /**
     * client event tells a buddy's typing status
     *
     * @param buddy who the status refers to
     * @param status 3 possible numbers, refer to the TOCConstants
     *
     * @see TOCConstants#CLIENT_EVENT_STOPPED
     * @see TOCConstants#CLIENT_EVENT_ENTERED
     * @see TOCConstants#CLIENT_EVENT_TYPING
     */
    void clientEvent(TOCBuddy buddy, int status);

    /**
     * Info about a buddy's icon?
     * @param buddy this response pertains to
     * @param unknown string with a variable length 
     * of parameters that appear to be delimited by spaces
     */
    void buddyArt(TOCBuddy buddy, String unknown);
    
    /**
     * Buddy signed on alert
     * 
     * @param buddy
     */
    void buddySignedOn(TOCBuddy buddy);
    
    /**
     * Buddy signed off alert
     * 
     * @param buddy
     */
    void buddySignedOff(TOCBuddy buddy);
    
    /**
     * Buddy is now away alert
     * 
     * @param buddy
     */
    void buddyAway(TOCBuddy buddy);
    
    /**
     * Buddy returned from away status
     * 
     * @param buddy
     */
    void buddyReturned(TOCBuddy buddy);
    
    /**
     * Buddy was added successfully
     * 
     * @param buddy
     */
    void buddyAdded(TOCBuddy buddy);
    
    /**
     * Error usually happens with bad command parameters, 
     * but could sometimes be the server's fault.
     * 
     * @param error the error code
     * @param args optional arguments to go with this error
     */
    void aimError(int error, String[] args);
    
    /**
     * Client joined a chatroom
     * 
     * @param chatroomID id to indentify this chatroom. Needed for other chat commands.
     * @param chatroomName Name of this chatroom
     */
    void chatroomJoined(String chatroomID, String chatroomName);
    
    /**
     * Message was received in this chatroom
     * DEPRECATED
     * @param chatroomID ID for this chatroom
     * @param sender name of who sent this message
     * @param message the message, including html tags
     * @param whisper true if only this client can see the message, false if its a public message.
     */
    void chatroomMessageReceived(String chatroomID, String sender, String message, boolean whisper);
    
    /**
     * Message was received in this chatroom
     * 
     * @param chatroomID ID for this chatroom
     * @param sender name of who sent this message
     * @param message the message, including html tags
     * @param whisper true if only this client can see the message, false if its a public message.
     * @param language
     * @param unknown 
     */
    void chatroomMessageReceived(String chatroomID, String sender, String message, boolean whisper, String language, String unknown);
    
    /**
     * When users enter or leave a chatroom
     * 
     * @param chatroomID ID for this chatroom
     * @param buddy name of person whose status changed
     * @param inside true if person joined room, false if left the room
     */
    void chatroomUpdate(String chatroomID, String buddy, boolean inside);
    
    /**
     * Client is being invited to join a chatroom
     * 
     * @param chatroomID ID for this chatroom
     * @param chatroomName name for this chatroom
     * @param sender the person who sent the invite
     * @param message optional greeting message
     */
    void chatroomInvite(String chatroomID, String chatroomName, String sender, String message);
    
    /**
     * Client was dropped from chatroom, either kicked, connection dropped, or left
     * 
     * @param chatroomID ID of chatroom the client left
     */
    void chatroomDropped(String chatroomID);
    
    /**
     * Event received when client attempts to change the format of their name
     * 
     * @param status status code, 0 mean success
     * @param args optional arguments if status is not 0
     */
    void adminFormatStatus(int status, String[] args);
    
    /**
     * Event received when client attempts to change their password
     * 
     * @param status status code, 0 mean success
     * @param args optional arguments if status is not 0
     */
    void adminPasswordStatus(int status, String[] args);
    
    /**
     * Event received when client requests a user's profile info or directory info
     * Use the server address and port the client connected to, followed by the
     * directory name passed here.
     * 
     * @param windowName suggested window name to use
     * @param url a directory name
     */
    void urlReceived(String windowName, String url);
    
    /**
     * Buddy list was received from the server.
     * This usually happens once just after connection.
     * This list is of all the buddies on a client's buddylist
     * 
     * @param buddies an Iterator of TOCBuddy objects
     */
    void buddyListReceived(Iterator buddies);
    
    /**
     * Event received when client attempts to set their directory info
     * 
     * @param status status code, 0 mean success
     * @param args optional arguments if status is not 0
     */
    void dirStatus(int status, String[] args);
    
    /**
     * Client has been warned
     * 
     * @param person who warned the client. If null, then warning was anonymous
     * @param amount the client's new warning percentage
     */
    void warned(TOCBuddy buddy, String amount);
    
    /**
     * For whatever reason, the server may choose to pause communication.
     * Any commands sent to the server at that point will be dropped, and
     * sometimes so will the connection.  Communication can resume again
     * once it is unpaused.
     * 
     * @param state true if communication is pause, false if not
     */
    void isPaused(boolean state);
    
    /**
     * Someone is proposing a rendezvous to perform a specific action.
     * This can be sending of buddy list, file transfer, etc...
     * 
     * @param buddy the person who sent the proposal
     * @param UUID action to perform. This is the same as the capabilities UUID's
     * @param cookie cookie used for other communication pertaining to this action
     * @param seq sequence number
     * @param rip IP address to rendezvous at (connect at)
     * @param pip proposer's IP address
     * @param vip proposer's verified IP address
     * @param port port to connect to at rip
     * @param tlv separated by colons and values are base64 encoded
     */
    void rvousPropose(TOCBuddy buddy, String UUID, String cookie, int seq,
                    String rip, String pip, String vip, String port, String tlv);

    
}
